from .small_view_set import SmallViewSet
from .config import SmallViewSetConfig
from .decorators import (
    endpoint,
    endpoint_disabled,
)
from.helpers import (
    default_exception_handler,
    default_options_and_head_handler,
)
from .exceptions import (
    BadRequest,
    EndpointDisabledException,
    MethodNotAllowed,
    Unauthorized,
)

__all__ = [
    "SmallViewSet",
    "SmallViewSetConfig",

    "endpoint",
    "endpoint_disabled",

    "default_exception_handler",
    "default_options_and_head_handler",

    "BadRequest",
    "EndpointDisabledException",
    "MethodNotAllowed",
    "Unauthorized",
]


import inspect
from django.conf import settings

from small_view_set.config import SmallViewSetConfig
from .exceptions import EndpointDisabledException

def endpoint(
        allowed_methods: list[str]):
    def decorator(func):
        func_name = func.__name__
        def sync_wrapper(viewset, *args, **kwargs):
            request = args[0]
            try:
                config: SmallViewSetConfig = getattr(settings, 'SMALL_VIEW_SET_CONFIG', SmallViewSetConfig())
                pre_response = config.options_and_head_handler(request, allowed_methods)
                if pre_response:
                    return pre_response
                pk = kwargs.pop('pk', None)
                if pk is None:
                    return func(viewset, request=request)
                else:
                    return func(viewset, request=request, pk=pk)
            except Exception as e:
                return config.exception_handler(request, func_name, e)

        async def async_wrapper(viewset, *args, **kwargs):
            request = args[0]
            try:
                config: SmallViewSetConfig = getattr(settings, 'SMALL_VIEW_SET_CONFIG', SmallViewSetConfig())
                pre_response = config.options_and_head_handler(request, allowed_methods)
                if pre_response:
                    return pre_response
                pk = kwargs.pop('pk', None)
                if pk is None:
                    return await func(viewset, request=request)
                else:
                    return await func(viewset, request=request, pk=pk)
            except Exception as e:
                return config.exception_handler(request, func_name, e)

        if inspect.iscoroutinefunction(func):
            return async_wrapper
        else:
            return sync_wrapper

    return decorator


def endpoint_disabled(func):
    """
    Temporarily disables an API endpoint based on the SMALL_VIEWSET_RESPECT_DISABLED_ENDPOINTS setting.

    When `SMALL_VIEWSET_RESPECT_DISABLED_ENDPOINTS` in Django settings is set to `True`, this decorator
    will raise an `EndpointDisabledException`, resulting in a 404 response. When set to `False`,
    the endpoint will remain active, which is useful for testing environments.

    Usage:
        - Apply this decorator directly to a view method or action.
        - Example:
        
        ```python
        class ExampleViewSet(SmallViewSet):

            @default_handle_endpoint_exceptions
            @endpoint_disabled
            def retrieve(self, request: Request) -> JsonResponse:
                self.protect_retrieve(request)
                . . .
        ```
    """
    config: SmallViewSetConfig = getattr(settings, 'SMALL_VIEW_SET_CONFIG', SmallViewSetConfig())
    def sync_wrapper(*args, **kwargs):
        if config.respect_disabled_endpoints:
            raise EndpointDisabledException()
        return func(*args, **kwargs)

    async def async_wrapper(*args, **kwargs):
        if config.respect_disabled_endpoints:
            raise EndpointDisabledException()
        return await func(*args, **kwargs)


    if inspect.iscoroutinefunction(func):
        return async_wrapper
    else:
        return sync_wrapper




import inspect
import json
import logging
from django.http import JsonResponse
from urllib.request import Request

from .exceptions import BadRequest, MethodNotAllowed

logger = logging.getLogger('app')

class SmallViewSet:
    def parse_json_body(self, request: Request):
        if request.content_type != 'application/json':
            raise BadRequest('Invalid content type')
        return json.loads(request.body)

    def protect_create(self, request: Request):
        """
        Stub for adding any custom business logic to protect the create method.
        For example:
        - Check if the user is authenticated
        - Check if the user has validated their email
        - Throttle requests

        Recommended to call super().protect_create(request) in the subclass in case
        this library adds logic in the future.
        """
        pass

    def protect_list(self, request: Request):
        """
        Stub for adding any custom business logic to protect the list method.
        For example:
        - Check if the user is authenticated
        - Check if the user has validated their email
        - Throttle requests

        Recommended to call super().protect_create(request) in the subclass in case
        this library adds logic in the future.
        """
        pass

    def protect_retrieve(self, request: Request):
        """
        Stub for adding any custom business logic to protect the retrieve method.
        For example:
        - Check if the user is authenticated
        - Check if the user has validated their email
        - Throttle requests

        Recommended to call super().protect_create(request) in the subclass in case
        this library adds logic in the future.
        """
        pass

    def protect_update(self, request: Request):
        """
        Stub for adding any custom business logic to protect the update method.
        For example:
        - Check if the user is authenticated
        - Check if the user has validated their email
        - Throttle requests

        Recommended to call super().protect_create(request) in the subclass in case
        this library adds logic in the future.
        """
        pass

    def protect_delete(self, request: Request):
        """
        Stub for adding any custom business logic to protect the delete method.
        For example:
        - Check if the user is authenticated
        - Check if the user has validated their email
        - Throttle requests

        Recommended to call super().protect_create(request) in the subclass in case
        this library adds logic in the future.
        """
        pass

    async def default_router(self, request: Request, pk=None, *args, **kwargs):
        """
        This method routes requests to the appropriate method based on the HTTP method and presence of a primary key (pk).
        
        It also handles errors and returns appropriate JSON responses by using the decorator @default_handle_endpoint_exceptions.
        
        GET/POST for collection endpoints and GET/PUT/PATCH/DELETE for detail endpoints.

        Example:
        ```
        # Note: AppViewSet is a subclass of SmallViewSet with overridden protect methods with more specific logic.

        class CommentViewSet(AppViewSet):
            def urlpatterns(self):
                return [
                    path('api/comments/',                     self.default_router, name='comments_collection'),
                    path('api/comments/<int:pk>/',            self.default_router, name='comments_detail'),
                    path('api/comments/<int:pk>/custom_put/', self.custom_put,     name='comments_custom_put_detail'),
                ]

            @default_handle_endpoint_exceptions
            def create(self, request: Request):
                self.protect_create(request)
                . . .

            @default_handle_endpoint_exceptions
            def update(self, request: Request, pk: int):
                self.protect_update(request)
                . . .

            @default_handle_endpoint_exceptions
            def custom_put(self, request: Request, pk: int):
                self.protect_update(request)
                . . .

            @disable_endpoint
            @default_handle_endpoint_exceptions
            def some_disabled_endpoint(self, request: Request):
                self.protect_retrieve(request)
                . . .
        ```
        """
        func = None
        if pk is None:
            if request.method == 'GET':
                if hasattr(self, 'list'):
                    func = self.list

            elif request.method == 'POST':
                if hasattr(self, 'create'):
                    func = self.create
        else:
            if request.method == 'GET':
                if hasattr(self, 'retrieve'):
                    func = self.retrieve

            elif request.method == 'PUT':
                if hasattr(self, 'put'):
                    func = self.put

            elif request.method == 'PATCH':
                if hasattr(self, 'patch'):
                    func = self.patch

            elif request.method == 'DELETE':
                if hasattr(self, 'delete'):
                    func = self.delete

        if func is None:
            raise MethodNotAllowed(request.method)

        if pk is not None:
            kwargs['pk'] = pk

        if inspect.iscoroutinefunction(func):
            return await func(request, *args, **kwargs)
        else:
            return func(request, *args, **kwargs)



from typing import Callable
from urllib.request import Request
from small_view_set.helpers import default_exception_handler, default_options_and_head_handler


class SmallViewSetConfig:
    """
    Configuration class for SmallViewSet.

    This class allows customization of exception handling and handling of
    OPTIONS and HEAD requests for endpoints.

    Args:
        exception_handler (Callable[[str, Exception], None]): A callback function
            for handling exceptions. The function takes two parameters:
            1. The name of the endpoint function (e.g., 'list', 'retrieve', or a custom endpoint name).
            2. The exception that was thrown.
        options_and_head_handler (Callable[[Request, list[str]], None]): A callback function
            for handling OPTIONS and HEAD requests. The function takes two parameters:
            1. The Django Request object.
            2. A list of allowed HTTP methods for the endpoint (e.g., ['PUT', 'PATCH']).
    """
    def __init__(
            self,
            exception_handler: Callable[[str, Exception], None] = default_exception_handler,
            options_and_head_handler: Callable[[Request, list[str]], None] = default_options_and_head_handler,
            respect_disabled_endpoints=True):
        self.exception_handler = exception_handler
        self.options_and_head_handler = options_and_head_handler
        self.respect_disabled_endpoints = respect_disabled_endpoints


import json
import logging

from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist, SuspiciousOperation, PermissionDenied
from django.http import Http404, JsonResponse
from urllib.request import Request

from .exceptions import EndpointDisabledException, MethodNotAllowed, Unauthorized


_logger = logging.getLogger('django-small-view-set.default_handle_endpoint_exceptions')
if not _logger.hasHandlers():
    handler = logging.StreamHandler()
    formatter = logging.Formatter('%(asctime)s %(levelname)s %(message)s')
    handler.setFormatter(formatter)
    _logger.addHandler(handler)
    _logger.setLevel(logging.INFO)


def default_options_and_head_handler(request, allowed_methods: list[str]):
    if request.method == 'OPTIONS':
            response = JsonResponse(
                data=None,
                safe=False,
                status=200,
                content_type='application/json')
            response['Allow'] = ', '.join(allowed_methods)
            return response

    if request.method == 'HEAD':
        response = JsonResponse(
            data=None,
            safe=False,
            status=200,
            content_type='application/json')
        response['Allow'] = ', '.join(allowed_methods)
        return response

    if request.method not in allowed_methods:
        raise MethodNotAllowed(method=request.method)


def default_exception_handler(request: Request, endpoint_name: str, exception):
    try:
        raise exception

    except json.JSONDecodeError:
        return JsonResponse(data={"errors": "Invalid JSON"}, status=400)

    except (TypeError, ValueError) as exception:
        if hasattr(exception, 'detail'):
            return JsonResponse(data={'errors': exception.detail}, status=400)
        if hasattr(exception, 'message'):
            return JsonResponse(data={'errors': exception.message}, status=400)
        return JsonResponse(data=None, safe=False, status=400)

    except Unauthorized:
        return JsonResponse(data=None, safe=False, status=401)

    except (PermissionDenied, SuspiciousOperation):
        return JsonResponse(data=None, safe=False, status=403)

    except (Http404, ObjectDoesNotExist):
        return JsonResponse(data=None, safe=False, status=404)
    
    except EndpointDisabledException:
        return JsonResponse(data=None, safe=False, status=405)

    except MethodNotAllowed as exception:
        return JsonResponse(
            data={'errors': f"Method {exception.method} is not allowed"},
            status=405)

    except Exception as exception:
        # Catch-all exception handler for API endpoints.
        # 
        # - Always defaults to HTTP 500 with "Internal server error" unless the exception provides a more specific status code and error details.
        # - Duck types to extract error information from `detail` or `message` attributes, if available.
        # - Never exposes internal exception contents to end users for 5xx server errors unless settings.DEBUG is True.
        # - Allows structured error payloads (string, list, or dict) without assumptions about the error format.
        # - Logs exceptions fully for server-side diagnostics, distinguishing handled vs unhandled cases.
        # 
        # This design prioritizes API security, developer debugging, and future portability across projects.

        status_code = getattr(exception, 'status_code', 500)
        error_contents = None

        if hasattr(exception, 'detail'):
            error_contents = exception.detail
        elif hasattr(exception, 'message') and isinstance(exception.message, str):
            error_contents = exception.message

        if 400 <= status_code <= 499:
            if status_code == 400:
                message = 'Bad request'
            elif status_code == 401:
                message = 'Unauthorized'
            elif status_code == 403:
                message = 'Forbidden'
            elif status_code == 404:
                message = 'Not found'
            elif status_code == 405:
                message = 'Method not allowed'
            elif status_code == 429:
                message = 'Too many requests'
            elif error_contents:
                message = error_contents
            else:
                message = 'An error occurred'

            if settings.DEBUG and error_contents:
                message = error_contents
        else:
            status_code = 500
            message = 'Internal server error'
            if settings.DEBUG:
                message = error_contents if error_contents else str(exception)

        e_name = type(exception).__name__
        if error_contents:
            msg = f"Handled API exception in {endpoint_name}: {e_name}: {error_contents}"
            _logger.error(msg)
                
        else:
            msg = f"Unhandled exception in {endpoint_name}: {e_name}: {exception}"
            _logger.error(msg)

        return JsonResponse(
            data={'errors': message},
            safe=False,
            status=status_code,
            content_type='application/json')


class EndpointDisabledException(Exception):
    status_code = 405
    message = "Method Not Allowed"
    error_code = "method_not_allowed"

class Unauthorized(Exception):
    status_code = 401
    message = "Unauthorized"
    error_code = "unauthorized"

class BadRequest(Exception):
    status_code = 400
    error_code = "bad_request"
    def __init__(self, message: str | list | dict):
        """
        Args:
            message (str | list | dict): 
                A JSON-serializable error description to send to the client. 
                Typically a string, list of errors, or dictionary of field errors.
        """
        self.message = message or "Bad Request"
        super().__init__(message)

class MethodNotAllowed(Exception):
    status_code = 405
    error_code = "method_not_allowed"
    def __init__(self, method: str):
        self.method = method
        self.message = f'Method {method} not allowed'
        super().__init__(self.message)